/* 
 * $Id: imonlcd.c,v 1.1 1999/06/15 21:01:11 lars Exp lars $
 * 
 * LCDproc ISDN line monitor client
 * includes code stolen from isdn4linux/imontty.c by Volker Gtz 
 *
 * Copyright (c) 1999 Lars Wessels <lars@nordsee.inka.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 *
 * $Log: imonlcd.c,v $
 * Revision 1.1  1999/06/15 21:01:11  lars
 * - implemented full client/server handshaking
 * - rewrote main update loop
 * - use select() to detect changes in /dev/isdninfo
 * - new command line option for verbose output
 * - added debug() messages, define DEBUG to activate
 *
 * Revision 1.0  1999/03/03 20:30:02  lars
 * Initial revision
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <linux/isdn.h>
#include "sockets.h"
#include "imonlcd.h"
#include "debug.h"

FILE *isdninfo;
int sock=0;
int verbose=0;
channel_info info[ISDN_MAX_CHANNELS];


void usage() {
	printf("%s v%s - (c) 1999 Lars Wessels <lars@nordsee.inka.de>\n",
		CLIENTNAME, VERSION);
	fprintf(stderr,"Usage: imond [-s server] [-p port] [-v]\n");
	exit(0);
}

void scan_str(char *buffer, char *field[], int max) {
    
	char *buf = buffer;
	int i;
	char *token = strtok(buf, "\t"); 

	for(i=0; i < max; i++) {
		token = strtok(NULL, " ");
		if (token == NULL) continue;
		if ((field[i] = malloc(strlen(token) * sizeof(char))) != NULL)
			strcpy(field[i], token);
	}
}

void scan_int(char *buffer, int (*field)[], int max) {
    
	char *buf = buffer;
	int i;
	char *token = strtok(buf, "\t");

	for(i=0; i < max; i++) {
		token = strtok(NULL, " ");
		if (token == NULL) continue;
		(*field)[i] = atoi(token);
	}
}

void clean_exit (int val) {
    sock_close(sock);
    fclose(isdninfo);
    exit(0);
}

void alarm_exit (int val) {
	fprintf(stderr,"No answer from LCDproc server. Exiting.\n");
   	clean_exit(0);
}

/* get channel info from /dev/isdninfo, inspired by imontty.c */
int read_isdn_info (FILE *isdninfo, channel_info *info) {
    
    char buf[BUFSIZE];
    char *idmap[ISDN_MAX_CHANNELS];
    int chmap[ISDN_MAX_CHANNELS];
    int drmap[ISDN_MAX_CHANNELS];
    int usage[ISDN_MAX_CHANNELS];
    int flags[ISDN_MAX_DRIVERS];
    char *phone[ISDN_MAX_CHANNELS];
    int i, active=0;


    fgets(buf, BUFSIZE, isdninfo);
    scan_str(buf, idmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &chmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &drmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &usage, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &flags, ISDN_MAX_DRIVERS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_str(buf, phone, ISDN_MAX_CHANNELS);


    for(i=0; i < ISDN_MAX_CHANNELS; i++, info++) {
		if (chmap[i] >= 0 && drmap[i] >= 0) {
			active++;

			info->chmap = chmap[i]; 
			info->idmap = idmap[i]; 
          
			/* values for usage inspired by display's character set */
			if (usage[i] & ISDN_USAGE_OUTGOING)
				info->usage = 126;  /* char value for "->" */
			else if (usage[i] == ISDN_USAGE_EXCLUSIVE)
				info->usage = 69;   /* char value for "E"  */
			else if (usage[i])
				info->usage = 127;  /* char value for "<-" */
			else
				info->usage = 79;   /* char value for "O"  */

			switch(usage[i] & ISDN_USAGE_MASK) {
				case ISDN_USAGE_RAW:
					info->type = "Raw"; break;
				case ISDN_USAGE_MODEM:
					info->type = "Modem"; break;
				case ISDN_USAGE_NET:
					info->type = "Net"; break;
				case ISDN_USAGE_VOICE:
					info->type = "Voice"; break;
				case ISDN_USAGE_FAX:
					info->type = "Fax"; break;
				default:
					info->type = "?"; break;
			}
		    
			if(usage[i] & ISDN_USAGE_MASK)
				info->phone = phone[i];
		} 
	}
	return active;
}

int check_isdn_status (FILE *isdninfo, channel_info *info) {
  
	struct timeval timeout;
	fd_set fdset;

	FD_ZERO(&fdset);
	FD_SET(fileno(isdninfo),&fdset);
	timeout.tv_sec =  0;
	timeout.tv_usec = 250;

	switch (select(32,&fdset,(fd_set *)0,(fd_set *)0,&timeout)) {
		case 0: /* nothing changed */
			return 0;    
		case 1:
			read_isdn_info(isdninfo, info); /* read new ISDN info */
			return 1;    
		default:
			fprintf(stderr,"ERROR: select(%s): %s\n", 
					PATH_ISDNINFO, strerror(errno));
			clean_exit(1);
	  }
}

/* setup initial ISDN line monitor screen
   returns number of active ISDN channels  */
int init_screen (void) {

	int i, active;
	char tmp[MAXMSG];
	channel_info *ch;

	sprintf(tmp, "client_set name {%s}\n",CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "client_set name {%s}\n",CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "screen_add %s\n",CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "screen_set %s name {%s}\n",CLIENTNAME,CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "widget_add %s title title\n",CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "widget_set %s title {%s}\n",CLIENTNAME,SCREENTITLE);
	sock_send_string(sock, tmp);

	/* get initial info about available ISDN channels */
	active = read_isdn_info(isdninfo, info);
	if (verbose)
		printf("monitoring %d ISDN channels\n", active);

	/* add string widget for each active channel */
	for (ch=info,i=0; i<active; ch++,i++) {
		if (ch->idmap != NULL) {
			sprintf(tmp, "widget_add %s ch%d string\n",CLIENTNAME,i);
			sock_send_string(sock, tmp);
		}	
	}
	return active;
}

int parse_server_reply(char *buffer, char *field[]) {
	
	char *reply_token, *cp; 
	int i,tokens;

	for(i=0,tokens=0; i < MAXTOKENS; i++) {
		if (i==0) reply_token = strtok(buffer, " ");
		else reply_token = strtok(NULL, " ");
		if (reply_token == NULL) continue;
		if ((field[i] = malloc(strlen(reply_token) * sizeof(char))) != NULL) {
			if ((cp = strchr(reply_token, '\n')) != NULL) *cp = '\0';
			strcpy(field[i], reply_token);
			tokens++;
		}
	}
	return tokens;
}

int main(int argc, char **argv) {
    
	char server[BUFSIZE] = "localhost";
	int port = LCDPORT;
	int i, j, p, x, tokens, ch_status, ch_status_prev=0,
		len=0, ch_active=0, sendinfo=0;
	channel_info *ch;
	char tmp[MAXMSG], buf[MAXMSG], *reply[MAXTOKENS];

	/* install signal handlers */
	signal(SIGINT, clean_exit);
	signal(SIGHUP, clean_exit);
	signal(SIGTERM, clean_exit);
	signal(SIGALRM, alarm_exit);

    /* parse command line */
	for(i=1, j=0; i<argc; i++) {
		if(argv[i][0] == '-') {
			switch(argv[i][1]) {
				/* server name */
				case 'S':
				case 's': if(argc < i+1) usage();
					strcpy(server, argv[++i]);
					 break;
				/* port */
				case 'P':
				case 'p': if(argc < i+1) usage();
					port = atoi(argv[++i]);
					break;
				/* verbose outpu */
				case 'V':
				case 'v': verbose=1;
					break;
				default: usage();
					break;
	    	}
		}
	}
	if (verbose)
		printf("%s v%s - (c) 1999 Lars Wessels <lars@nordsee.inka.de>\n",
			CLIENTNAME, VERSION);
    
	if (!(isdninfo = fopen(PATH_ISDNINFO, "r"))) {
		fprintf(stderr, "ERROR: can't open '%s'\n",
				PATH_ISDNINFO, strerror(errno));
		exit(1);
	}

	/* connect to the LCDd on server */
	sock = sock_connect(server, port);
	if (sock <= 0) {
		fprintf(stderr,"ERROR: can't connect to server %s on port %i.\n",
			 server, port);
		clean_exit(1);
	}

	/* initial client->server handshake with timeout */
	sock_send_string(sock, "hello\n");
	alarm(15);
	while (len <= 0) {
		len = sock_recv(sock, buf, MAXMSG);
	}
	alarm(0);
	debug("socket in: %s\n", buf);

	/* is there an LCDproc server on the other end? */
	tokens = parse_server_reply(buf, reply);
	if (tokens > 0 && !strcmp(reply[0], "connect")) {
		if (verbose)
			printf("connected to LCDproc server %s\n", reply[2]);
	} else {
		fprintf(stderr,"ERROR: handshaking with server failed:\n");
		fprintf(stderr,"expecting \"connect\" ");
		fprintf(stderr,"but server says \"%s\"\n", buf);
		clean_exit(1);
	}	
	memset(&buf,0,sizeof(buf));

	/* init screen widgets and get number of active ISDN channels */
	ch_active = init_screen();	

	/* main update loop */
	while(1) { 

		/* pay attention to server commands */
		len = sock_recv(sock, buf, MAXMSG);
		if (len > 0) {
			debug("socket in: %s\n", buf);
			tokens = parse_server_reply(buf, reply);
			memset(&buf,0,sizeof(buf));

			/* is he talking to us ? */
			if (tokens > 1 && !strcmp(reply[1], CLIENTNAME)) 
				sendinfo = strcmp(reply[0], "listen") ? 0 : 1;

			/* server terminates */
			if (!strcmp(reply[0], "bye")) {
				fprintf(stderr,"LCDproc server said 'bye'. Exiting.\n");
				clean_exit(0);
			}
		} else if (!len) {
			fprintf(stderr,"LCDproc probably died. Exiting.\n");
			clean_exit(0);	
		}

		ch_status = 0; 
		if (check_isdn_status(isdninfo, info) && verbose)
			printf("%s changed\n", PATH_ISDNINFO);
		for (ch=info,i=0; i<ch_active; ch++,i++) {
			if (ch->usage == 79) {  /* char value for "O" (off) */
				sprintf(tmp, "widget_set %s ch%d 1 %d {%d OFFLINE}\n",
					CLIENTNAME,i,(i%2)+2,ch->chmap);
			} else {
				ch_status++; 
				sprintf(tmp, "widget_set %s ch%d 1 %d {%d %s %c %s}\n",
					CLIENTNAME,i,(i%2)+2,i,ch->type,ch->usage,ch->phone);

				if (verbose && ch_status != ch_status_prev) {
					if (ch->usage == 126) x = 64; 
					else if (ch->usage == 127) x = 67;
					printf("%d %s %c %s\n",i,ch->type,ch->usage-x,ch->phone);
				}
			}
			if (sendinfo) {
				debug("socket out: %s\n", tmp);
				sock_send_string(sock, tmp);
			}
			/* show max. 2 channels at a time ... rough */
			if (ch_active > 2 && i%2 == 1) sleep(1); 
		} 

		/* screen priority is raised if a channel is busy */
		p = (ch_status > 0) ? PRIORITY : 128;
		if (ch_status != ch_status_prev) {
			if (verbose)
				printf("setting screen priority to %d\n", p);
			sprintf(tmp, "screen_set %s priority %d\n",CLIENTNAME,p);
			debug("socket out: %s\n",tmp);
			sock_send_string(sock, tmp); 
			ch_status_prev = ch_status;
		}
		usleep(100000); 
	} 
}
